#
# Copyright (c) 2020-2021 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
import decimal
import json

from .register_row import RegisterRow


def sensor_port_name(index):
    """Sensor port name for sensor port with index INDEX, which must be
    in the range from 0..SENSOR_PORT_COUNT-1.

    """
    return "S%u" % (index + 1)


class Local:
    """Fetch data from /api/local."""

    # Sensor parameter spec strings:
    #   For l=...:
    SPEC_L1 = "L1"
    SPEC_L2 = "L2"
    SPEC_L3 = "L3"
    SPEC_L1_L2 = "L12"
    SPEC_L2_L3 = "L23"
    SPEC_L3_L1 = "L31"
    SPEC_LDC = "Ldc"

    # Sensor names as they appear in the output:
    NAME_L1 = "L1"
    NAME_L2 = "L2"
    NAME_L3 = "L3"
    NAME_L1_L2 = "L1-L2"
    NAME_L2_L3 = "L2-L3"
    NAME_L3_L1 = "L3-L1"
    NAME_LDC = "Ldc"
    NAME_TEMP_PCB = "Tpcb"
    NAME_HUMID_PCB = "Hpcb"

    METRIC_RATE = "rate"
    METRIC_CUMUL = "cumul"
    METRIC_TYPE = "type"

    MEASUREMENT_NORMAL = "n"
    MEASUREMENT_MEAN = "m"
    MEASUREMENT_FREQ = "f"

    SECTION_VALUES = "values"
    SECTION_ENERGY = "energy"
    SECTION_APPARENT = "apparent"

    def __init__(self, dev, params, **kwargs):
        self.raw = dev.get("/local", params=params, **kwargs)

    def ts(self):
        """Return the timestamp of the local data as a Decimal()."""
        if self.raw is None:
            return None
        return decimal.Decimal(self.raw["ts"])

    def type_code(self, sensor_name):
        """Return the type-code of sensor SENSOR_NAME."""
        if self.raw is None:
            return None
        try:
            return self.raw["values"][sensor_name]["type"]
        except KeyError:
            pass
        return None

    def rate(
        self,
        sensor_name,
        measurement=MEASUREMENT_NORMAL,
        section=SECTION_VALUES,
    ):
        """Return the rate of change value of sensor SENSOR_NAME.  By default,
        the normal value for the sensor is returned.  The mean value
        or frequency can be requested by passing MEASUREMENT_MEAN or
        MEASUREMENT_FREQ, respectively, as MEASUREMENT.  If the value
        is not available, None is returned.

        """
        if self.raw is None:
            return None
        try:
            m = self.raw[section][sensor_name]["rate"]
            if section == Local.SECTION_VALUES:
                return m[measurement]
            return m
        except KeyError:
            pass
        return None

    def cumul(
        self,
        sensor_name,
        measurement=MEASUREMENT_NORMAL,
        section=SECTION_VALUES,
    ):
        """Return the cumulative value of sensor SENSOR_NAME.  By default, the
        normal value for the sensor is returned.  The mean value or
        frequency can be requested by passing MEASUREMENT_MEAN or
        MEASUREMENT_FREQ, respectively, as MEASUREMENT.  If the value
        is not available, None is returned.

        """
        if self.raw is None:
            return None
        try:
            m = self.raw[section][sensor_name]["cumul"]
            if section == Local.SECTION_VALUES:
                return int(m[measurement])
            return int(m)
        except KeyError:
            pass
        return None

    def row(
        self,
        metric=METRIC_CUMUL,
        measurement=MEASUREMENT_NORMAL,
        section=SECTION_VALUES,
    ):
        """Return a timestamped row of register data.  By default, the
        cumulative normal values of section "values" is returned.
        Non-default rows can be selected by passing the desired values
        for METRIC, MEASUREMENT, and/or SECTION.

        """
        if self.raw is None:
            return None
        section = self.raw[section]
        regs = {}
        for key, metrics in section.items():
            if metric == Local.METRIC_TYPE:
                val = metrics[metric]
            else:
                val = metrics[metric]
                if section == Local.SECTION_VALUES:
                    val = val[measurement]
                if metric == Local.METRIC_CUMUL:
                    val = int(val)  # convert from decimal string to integer
            regs[key] = val
        return RegisterRow(self.ts(), regs)

    def __str__(self):
        return json.dumps(self.raw)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#  All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
class BitStuffer:
    def __init__(self, header=b""):
        self.run_length = 0  # current length of run of 1 bits
        self.output = header
        self.out_byte = 0x00
        self.out_mask = 0x80

    def shift(self):
        self.out_mask >>= 1
        if self.out_mask == 0:
            self.output += bytes((self.out_byte,))
            self.out_byte = 0x00
            self.out_mask = 0x80

    def append(self, byte):
        mask = 0x80
        while mask != 0:
            if byte & mask:
                self.out_byte |= self.out_mask
                self.run_length += 1
                if self.run_length >= 7:
                    self.shift()
                    self.run_length = 0
            else:
                self.run_length = 0
            self.shift()
            mask >>= 1

    def finish(self):
        while self.out_mask != 0x80:
            self.shift()

    def get_output(self):
        self.finish()
        return self.output



#
# Copyright (c) 2020-2022 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""This module provides support additional requests auth services, in
particular for JWT-token based authentication (JWTAuth) and for plain
token-based authentication (TokenAuth)."""

import hashlib
import os
import secrets
import types

from functools import wraps
from urllib.parse import urlparse

import requests

from . import json_api


# The name of the optional environment variable storing a token:
ENV_EGAUGE_API_TOKEN = "EGAUGE_API_TOKEN"


def _decorate_public_metaclass(decorator):
    """Return a metaclass which will decorate all public methods of a
    class with the DECORATOR function.

    """

    class MetaClass(type):
        def __new__(mcs, class_name, bases, class_dict, **kwargs):
            if bases:
                decorated_class = bases[0]
                for attr_name, attr in decorated_class.__dict__.items():
                    if isinstance(attr, types.FunctionType):
                        if attr_name[0] == "_":
                            continue
                        attr = decorator(attr)
                    class_dict[attr_name] = attr
            return type.__new__(mcs, class_name, bases, class_dict, **kwargs)

    return MetaClass


def decorate_public(cls, decorator):
    """Return a subclass of CLS in which all public methods of CLS are
    decorated with function DECORATOR.  Methods whose name start with
    an underscore ('_') are considered private, all other methods are
    considered public.

    """

    # pylint: disable=unused-variable
    def wrapper(method):
        @wraps(method)
        def wrapped(*args, **kwargs):
            return decorator(method, *args, *kwargs)

        return wrapped

    class DecoratedClass(cls, metaclass=_decorate_public_metaclass(wrapper)):
        # pylint: disable=too-few-public-methods
        pass

    return DecoratedClass


class JWTAuth(requests.auth.AuthBase):
    """Implements the eGauge device WebAPI's JWT-based authentication
    scheme.  Digest login is used so the password is never sent over
    the HTTP connection.

    """

    def __init__(self, username, password):
        self.bearer_token = None
        self.username = username
        self.password = password

    def __call__(self, r):
        if self.bearer_token:
            r.headers["Authorization"] = self.auth_header()
        r.register_hook("response", self.handle_401)
        return r

    def __eq__(self, other):
        return self.username == getattr(
            other, "username", None
        ) and self.password == getattr(other, "password", None)

    def auth_header(self):
        """Return HTTP Auth-Header value."""
        return "Bearer " + self.bearer_token

    def handle_401(self, r, **kwargs):
        """Called when server responds with 401 Unauthorized."""
        if r.status_code != 401:
            return r

        try:
            auth_request = r.json()
        except ValueError:
            return r

        realm = auth_request["rlm"]
        server_nonce = auth_request["nnc"]

        client_nonce = f"{secrets.randbits(64):x}"

        content = self.username + ":" + realm + ":" + self.password
        ha1 = hashlib.md5(content.encode("utf-8")).hexdigest()

        content = ha1 + ":" + server_nonce + ":" + client_nonce
        ha2 = hashlib.md5(content.encode("utf-8")).hexdigest()

        data = {
            "rlm": realm,
            "usr": self.username,
            "nnc": server_nonce,
            "cnnc": client_nonce,
            "hash": ha2,
        }

        url = urlparse(r.request.url)
        login_uri = url.scheme + "://" + url.netloc + "/api/auth/login"
        verify = kwargs.get("verify", True)
        auth_reply = json_api.post(login_uri, data, timeout=60, verify=verify)

        if auth_reply is None:
            return r

        if "jwt" not in auth_reply:
            return r

        self.bearer_token = auth_reply["jwt"]

        prep = r.request.copy()
        prep.headers["Authorization"] = self.auth_header()
        _r = r.connection.send(prep, **kwargs)
        _r.history.append(r)
        _r.request = prep
        return _r


class TokenAuth(requests.auth.AuthBase):
    """Implements the eGauge web services' token-based authentication
    scheme.  This sends the password to the server, so it must not be
    used unless the underlying transport is encrypted!

    """

    def __init__(
        self,
        username=None,
        password=None,
        ask=None,
        token_service_url="https://api.egauge.net/v1/api-token-auth/",
    ):
        self.username = username
        self.password = password
        self.ask_credentials = ask
        self.token_file = None
        self.token_service_url = token_service_url
        self.token = os.environ.get(ENV_EGAUGE_API_TOKEN)
        if self.token is None:
            self.token_file = os.path.join(
                os.path.join(os.getenv("HOME"), ".egauge_api_token")
            )
            try:
                with open(self.token_file, "r", encoding="utf-8") as f:
                    self.token = f.read().rstrip()
                    if len(self.token) < 32:
                        self.token = None
            except IOError:
                pass

    def __call__(self, r):
        if self.token:
            r.headers["Authorization"] = self.auth_header()
        r.register_hook("response", self.handle_401)
        return r

    def __eq__(self, other):
        return self.username == getattr(
            other, "username", None
        ) and self.password == getattr(other, "password", None)

    def auth_header(self):
        """Return HTTP Auth-Header value."""
        return "Token " + self.token

    def handle_401(self, r, **kwargs):
        """Called when server responds with 401 Unauthorized."""
        if r.status_code != 401:
            return r

        usr = self.username
        pwd = self.password

        if usr is None or pwd is None:
            credentials = self.ask_credentials()
            if credentials is None:
                return r
            [usr, pwd] = credentials

        creds = {"username": usr, "password": pwd}
        verify = kwargs.get("verify", True)
        auth_reply = requests.post(
            self.token_service_url, json=creds, timeout=60, verify=verify
        ).json()

        if auth_reply is None:
            return r

        if "token" not in auth_reply:
            return r

        self.token = auth_reply["token"]

        if self.token_file is None:
            # the original token came for the os.environ
            os.environ[ENV_EGAUGE_API_TOKEN] = self.token
        else:
            try:
                fd = os.open(self.token_file, os.O_CREAT | os.O_WRONLY, 0o600)
                os.write(fd, (self.token + "\n").encode("utf-8"))
                os.close(fd)
            except IOError:
                pass

        prep = r.request.copy()
        prep.headers["Authorization"] = self.auth_header()
        _r = r.connection.send(prep, **kwargs)
        _r.history.append(r)
        _r.request = prep
        return _r



#
# Copyright (c) 2020 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""This module provides access to the eGauge WebAPI's /api/register
service."""

import decimal
import json

from ..error import Error
from .register_row import RegisterRow


class RegisterError(Error):
    """Raised if for any register related errors."""


def regnames_to_ranges(dev, regs):
    """Convert a list of register names to a register-ranges string."""
    indices = []
    for name in regs:
        indices.append(dev.reg_idx(name))
    indices = sorted(indices)

    ranges = []
    idx = 0
    while idx < len(indices):
        start = stop = indices[idx]
        idx += 1
        while idx < len(indices) and indices[idx] == stop + 1:
            stop = indices[idx]
            idx += 1
        if start == stop:
            ranges.append("%s" % start)
        else:
            ranges.append("%s:%s" % (start, stop))
    if len(ranges) < 1:
        return None
    return "+".join(ranges)


class Register:
    def __init__(self, dev, params=None, regs=None, **kwargs):
        """Fetch register data from device DEV.  PARAMS is a dictionary
        of query parameters that specify the data to return.  If REGS
        is none, all registers are returned.  If not None, it must be
        a list of register names for which data should be returned.
        Additional keyword arguments are passed along the
        requests.get().

        """
        self.dev = dev
        self.regorder = None
        self.iter_range_idx = None
        self.iter_row_idx = None
        self.iter_ts = None
        if params is None:
            params = {}
        if regs is not None:
            reg_ranges = regnames_to_ranges(dev, regs)
            if reg_ranges is not None:
                params["reg"] = reg_ranges
        self.raw = self.dev.get("/register", params=params, **kwargs)
        if self.raw is None:
            raise RegisterError("Failed to read register data.", params)
        if "error" in self.raw:
            raise RegisterError(
                "Error reading register data.", self.raw["error"]
            )

    def _create_regorder(self):
        self.regorder = {}
        for index, reg in enumerate(self.raw["registers"]):
            self.regorder[reg["name"]] = index

    def ts(self):
        """Return the timestamp of the register data as a Decimal()."""
        if self.raw is None:
            return None
        return decimal.Decimal(self.raw["ts"])

    def have(self, reg):
        """Return True if register REG exists, False otherwise."""
        if self.regorder is None:
            self._create_regorder()
        return reg in self.regorder

    def rate(self, reg):
        """Return the rate of change value of the register REG."""
        if self.raw is None:
            return None
        if self.regorder is None:
            self._create_regorder()
        return self.raw["registers"][self.regorder[reg]]["rate"]

    def type_code(self, reg):
        """Return the type-code of register REG."""
        if self.raw is None:
            return None
        if self.regorder is None:
            self._create_regorder()
        return self.raw["registers"][self.regorder[reg]]["type"]

    def formula(self, reg):
        """Return the formula of virtual register REG."""
        if self.raw is None:
            return None
        if self.regorder is None:
            self._create_regorder()
        return self.raw["registers"][self.regorder[reg]]["formula"]

    def index(self, reg):
        """Return the register index of register REG."""
        if self.raw is None:
            return None
        if self.regorder is None:
            self._create_regorder()
        return self.raw["registers"][self.regorder[reg]]["idx"]

    def database_id(self, reg):
        """Return the database-ID of register REG."""
        if self.raw is None:
            return None
        if self.regorder is None:
            self._create_regorder()
        return self.raw["registers"][self.regorder[reg]]["did"]

    def _make_row(self, ts, cumul_list):
        row = RegisterRow(ts)
        for idx, cumul in enumerate(cumul_list):
            name = self.raw["registers"][idx]["name"]
            row.regs[name] = int(cumul)
        return row

    def __getitem__(self, index):
        """Return the n-th row of cumulative register values as a
        RegisterRow.

        """
        range_idx = 0
        while True:
            curr_range = self.raw["ranges"][range_idx]
            ts = decimal.Decimal(curr_range["ts"])
            if index < len(curr_range["rows"]):
                ts -= index * decimal.Decimal(curr_range["delta"])
                return self._make_row(ts, curr_range["rows"][index])
            range_idx += 1
            index -= len(curr_range["rows"])

    def __iter__(self):
        """Iterate over cumulative register values.  Each iteration returns
        one row of data as a RegisterRow.

        """
        if self.raw is None:
            return None
        self.iter_range_idx = 0
        self.iter_row_idx = 0
        return self

    def __next__(self):
        if (
            self.raw is None
            or "ranges" not in self.raw
            or self.iter_range_idx >= len(self.raw["ranges"])
        ):
            raise StopIteration
        curr_range = self.raw["ranges"][self.iter_range_idx]

        if self.iter_row_idx == 0:
            self.iter_ts = decimal.Decimal(curr_range["ts"])

        row = self._make_row(
            self.iter_ts, curr_range["rows"][self.iter_row_idx]
        )

        self.iter_ts -= decimal.Decimal(curr_range["delta"])
        self.iter_row_idx += 1
        if self.iter_row_idx >= len(curr_range["rows"]):
            self.iter_row_idx = 0
            self.iter_range_idx += 1
        return row

    def __str__(self):
        return json.dumps(self.raw)



#
# Copyright (c) 2020-2021 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""This module provides access to the eGauge WebAPI's /api/ctid
service."""

import datetime
import os
import secrets
import time

from egauge import ctid
from egauge import webapi

from ..error import Error

SCAN_TIMEOUT = 2.5  # scan timeout in seconds


def ctid_info_to_table(reply):
    """Convert a ctid service REPLY to a CTid table."""
    t = ctid.Table()
    t.version = reply.get("version")
    t.mfg_id = reply.get("mfgid")
    t.model = reply.get("model")
    t.serial_number = reply.get("sn")
    t.sensor_type = reply.get("k")
    t.r_source = reply.get("rsrc")
    t.r_load = reply.get("rload")

    params = reply.get("params", {})
    t.size = params.get("size")
    t.rated_current = params.get("i")
    t.voltage_at_rated_current = params.get("v")
    t.phase_at_rated_current = params.get("a")
    t.voltage_temp_coeff = params.get("tv")
    t.phase_temp_coeff = params.get("ta")
    t.cal_table = {}
    cal_table = params.get("cal", {})
    for l_str in cal_table:
        l = float(l_str)
        t.cal_table[l] = [
            cal_table[l_str].get("v", 0),
            cal_table[l_str].get("a", 0),
        ]
    t.bias_voltage = params.get("bias_voltage", 0)
    t.scale = params.get("scale")
    t.offset = params.get("offset")
    t.delay = params.get("delay")
    t.threshold = params.get("threshold")
    t.hysteresis = params.get("hysteresis")
    t.debouce_time = params.get("debounce_time")
    t.edge_mask = params.get("edge_mask")
    t.ntc_a = params.get("ntc_a")
    t.ntc_b = params.get("ntc_b")
    t.ntc_c = params.get("ntc_c")
    t.ntc_m = params.get("ntc_m")
    t.ntc_n = params.get("ntc_n")
    t.ntc_k = params.get("ntc_k")
    return t


class CTidInfoError(Error):
    """Raised if for any CTid info related errors."""


class PortInfo:
    """Encapsulates the port number on which a CTid table was read, the
    polarity which is was read with, and the table itself.

    """

    def __init__(self, port, polarity, table):
        self.port = port
        self.polarity = polarity
        self.table = table

    def port_name(self):
        """Return the canonical port name."""
        return "S%d" % self.port

    def short_mfg_name(self):
        """Return the short (concise) name of the manufacturer of the sensor
        or `-' if unknown.

        """
        if self.table is None or self.table.mfg_id is None:
            return "-"
        return ctid.mfg_short_name(self.table.mfg_id)

    def model_name(self):
        """Return the model name of the sensor attached to the port.  If
        unknown `-' is returned.

        """
        if self.table is None or self.table.model is None:
            return "-"
        return self.table.model

    def mfg_model_name(self):
        """Return a "fully qualified" model name, which consists of the short
        version of the manufacter's name, a dash, and the model
        name.

        """
        return "%s-%s" % (self.short_mfg_name(), self.model_name())

    def sn(self):
        """Return the serial number or None if unknown."""
        if self.table is None or self.table.serial_number is None:
            return None
        return self.table.serial_number

    def serial_number(self):
        """Return the serial number of the sensor attached to the port as a
        decimal string.  If unknown, '-' is returned.

        """
        if self.table is None or self.table.serial_number is None:
            return "-"
        return str(self.table.serial_number)

    def unique_name(self):
        """Return a sensor's unique name, which is a string consisting of the
        manufacturer's short name, the model name, and the serial
        number, all separated by dashes..

        """
        return "%s-%s" % (self.mfg_model_name(), self.serial_number())

    def sensor_type(self):
        """Return the sensor type of the sensor attached to the port or None
        if unknown.

        """
        if self.table is None or self.table.sensor_type is None:
            return None
        return self.table.sensor_type

    def sensor_type_name(self):
        """Return the name of the sensor type of the sensor attached to the
        port or '-' if unknown.

        """
        st = self.sensor_type()
        if st is None:
            return "-"
        return ctid.get_sensor_type_name(st)

    def as_dict(self):
        """Return CTid info as a serializable dictionary."""
        if self.table is None:
            return None
        params = {}
        p = {
            "port": self.port,
            "polarity": self.polarity,
            "version": self.table.version,
            "mfgid": self.table.mfg_id,
            "model": self.table.model,
            "sn": self.table.serial_number,
            "k": self.table.sensor_type,
            "rsrc": self.table.r_source,
            "rload": self.table.r_load,
            "params": params,
        }
        if self.table.sensor_type in [
            ctid.SENSOR_TYPE_AC,
            ctid.SENSOR_TYPE_DC,
            ctid.SENSOR_TYPE_RC,
        ]:
            params["size"] = self.table.size
            params["i"] = self.table.rated_current
            params["v"] = self.table.voltage_at_rated_current
            params["a"] = self.table.phase_at_rated_current
            params["tv"] = self.table.voltage_temp_coeff
            params["ta"] = self.table.phase_temp_coeff
            params["bias_voltage"] = self.table.bias_voltage
            cal_table = {}
            for l, row in self.table.cal_table.items():
                cal_table[l] = {"v": row[0], "a": row[1]}
            params["cal"] = cal_table
        elif self.table.sensor_type == ctid.SENSOR_TYPE_VOLTAGE:
            params["scale"] = self.table.scale
            params["offset"] = self.table.offset
            params["delay"] = self.table.delay
        elif self.table.sensor_type == ctid.SENSOR_TYPE_TEMP_LINEAR:
            params["scale"] = self.table.scale
            params["offset"] = self.table.offset
        elif self.table.sensor_type == ctid.SENSOR_TYPE_TEMP_NTC:
            params["ntc_a"] = self.table.ntc_a
            params["ntc_b"] = self.table.ntc_b
            params["ntc_c"] = self.table.ntc_c
            params["ntc_m"] = self.table.ntc_m
            params["ntc_n"] = self.table.ntc_n
            params["ntc_k"] = self.table.ntc_k
        elif self.table.sensor_type == ctid.SENSOR_TYPE_PULSE:
            params["threshold"] = self.table.threshold
            params["hysteresis"] = self.table.hysteresys
            params["debounce_time"] = self.table.debounce_time
            params["edge_mask"] = self.table.edge_mask
        return p

    def __str__(self):
        return "(port=%d,polarity=%s,table=%s)" % (
            self.port,
            self.polarity,
            self.table,
        )


class CTidInfo:
    def __init__(self, dev):
        """Create an object which can be used to access the CTid service of
        device DEV.  The service allows reading CTid info from a
        particular port, scan for such info, flash the attached
        sensor's indicator LED, or iterate over all the ports with
        CTid information.

        """
        self.dev = dev
        self.tid = None
        self.info = None
        self.index = None
        self.polarity = None
        self.port_number = None

    def _make_tid(self):
        """Create a random transaction ID."""
        self.tid = secrets.randbits(32)
        if self.tid < 1:
            self.tid += 1

    def stop(self):
        """Stop pending CTid operation, if any."""
        if self.tid is not None:
            self.dev.post("/ctid/stop", {})
        self.tid = None

    def scan_start(self, port_number, polarity):
        """Initiate a CTid scan of PORT_NUMBER with POLARITY.  POLARITY must
        be either "+" or "-".

        """
        if port_number < 1:
            raise CTidInfoError("Invalid port number.", port_number)
        if self.tid is not None:
            self.stop()
        self._make_tid()
        self.polarity = polarity
        self.port_number = port_number
        data = {"op": "scan", "tid": self.tid, "polarity": polarity}
        resource = "/ctid/%d" % port_number
        last_e = None
        for _ in range(3):
            try:
                reply = self.dev.post(resource, data)
                if reply.get("status") == "OK":
                    return
            except Error as e:
                last_e = e
        raise CTidInfoError(
            "Failed to initiate CTid scan", port_number, polarity
        ) from last_e

    def scan_result(self):
        """Attempt to read result from a CTid scan initiated with a call to
        scan_start().  If the result is not available, None is
        returned.  If None is returned, the caller should wait a
        little and then retry the request again for up to SCAN_TIMEOUT
        seconds.

        """
        resource = "/ctid/%d" % self.port_number
        reply = self.dev.get(resource, params={"tid": self.tid})
        if (
            reply.get("port") == self.port_number
            and reply.get("tid") == self.tid
        ):
            return PortInfo(
                self.port_number, self.polarity, ctid_info_to_table(reply)
            )
        return None

    def scan(self, port_number, polarity=None, timeout=SCAN_TIMEOUT):
        """Scan the CTid information from the sensor attached to the specified
        PORT_NUMBER and return a PortInfo object as a result.  If no
        CTid info could be read from the port, the returned object's
        table member will be None.  If POLARITY is unspecifed, normal
        (positive) polarity will be attempted first and, if that
        fails, negative polarity will be attempted second.  If
        POLARITY is specified, it should be '+' to scan with normal
        polarity only or '-' to scan with reversed polarity only.  If
        TIMEOUT is unspecified, the operation will time out after
        SCAN_TIMEOUT seconds.  Otherwise, TIMEOUT should be a positive
        number specifying the number of seconds after which to time
        out the operation.

        """
        polarity_list = ["+", "-"] if polarity is None else [polarity]
        for pol in polarity_list:
            self.scan_start(port_number, pol)

            start_time = datetime.datetime.now()
            while True:
                time.sleep(0.25)
                result = self.scan_result()
                if result is not None:
                    return result
                elapsed = (
                    datetime.datetime.now() - start_time
                ).total_seconds()
                if elapsed > timeout:
                    break
            self.stop()
        return PortInfo(port_number, None, None)

    def flash(self, port_number, polarity="-"):
        """Flash the indicator LED on the sensor attached to the specified
        PORT_NUMBER using the specified POLARITY.  Flashing will
        continue until stop() is called (or until a timeout occurs
        after about 30 minutes).

        """
        if port_number < 1:
            raise CTidInfoError("Invalid port number.", port_number)
        if self.tid is not None:
            self.stop()
        self._make_tid()
        data = {"op": "flash", "tid": self.tid, "polarity": polarity}
        resource = "/ctid/%d" % port_number
        for _ in range(3):
            try:
                reply = self.dev.post(resource, data)
                if reply.get("status") == "OK":
                    break
            except Error:
                pass

    def delete(self, port_number):
        """Delete the CTid information stored for the specified port
        number.

        """
        if port_number < 1:
            raise CTidInfoError("Invalid port number.", port_number)
        resource = "/ctid/%d" % port_number
        reply = self.dev.delete(resource)
        if reply is None or reply.get("status") != "OK":
            reason = reply.get("error") if reply is not None else "timed out"
            raise CTidInfoError(
                "Failed to delete CTid info.", port_number, reason
            )

    def get(self, port_number):
        """Get the CTid information stored for the specified PORT_NUMBER (if
        any).

        """
        if port_number < 1:
            raise CTidInfoError("Invalid port number.", port_number)
        resource = "/ctid/%d" % port_number
        reply = self.dev.get(resource)
        if reply is None:
            raise CTidInfoError("Failed to read CTid info.", port_number)
        if len(reply) == 0:
            return None
        if reply.get("port") != port_number:
            raise CTidInfoError(
                "CTid info has incorrect port number.",
                reply.get("port"),
                port_number,
            )
        return PortInfo(
            port_number, reply.get("polarity"), ctid_info_to_table(reply)
        )

    def put(self, port_info):
        """Replace the existing CTid info with the one specified by PORT_INFO,
        which may be a single PortInfo() object or an array of
        PortInfo() objects.

        """
        if isinstance(port_info, list):
            resource = "/ctid"
            data = {"info": [pi.as_dict() for pi in port_info]}
        else:
            resource = "/ctid/%d" % port_info["port"]
            data = port_info.as_dict()
        reply = self.dev.put(resource, json_data=data)
        if reply is None:
            raise CTidInfoError("PUT of CTid info failed.")
        if reply.get("status") != "OK":
            raise CTidInfoError("Failure saving CTid info.", data, reply)

    def __iter__(self):
        """Iterate over all available CTid information."""
        reply = self.dev.get("/ctid")
        self.info = reply.get("info", [])
        self.index = 0
        return self

    def __next__(self):
        if self.index >= len(self.info):
            raise StopIteration
        info = self.info[self.index]
        t = ctid_info_to_table(info)
        self.index += 1
        return PortInfo(info["port"], info["polarity"], t)


if __name__ == "__main__":
    from . import device

    dut = os.getenv("EGAUGE_DUT") or "http://1608050004.lan"
    usr = os.getenv("EGAUGE_USR") or "owner"
    pwd = os.getenv("EGAUGE_PWD") or "default"
    ctid_info = CTidInfo(device.Device(dut, auth=webapi.JWTAuth(usr, pwd)))
    print("SCANNING")
    port_info = ctid_info.scan(port_number=3)
    print("  port_info[%d]" % port_info.port, port_info.table)
    print("-" * 40)
    print("ITERATING")
    for t in ctid_info:
        print("  port %d%s:" % (t.port, t.polarity), t.table)

    print("DELETING")
    ctid_info.delete(port_number=3)
    port_info = ctid_info.get(port_number=3)
    if port_info is None:
        print("  no CTid info for port 3")
    else:
        print("  port_info[%d]" % port_info.port, port_info.table)

    print("FLASHING")
    ctid_info.flash(port_number=3)
    time.sleep(5)
    ctid_info.stop()



#
#   Copyright (c) 2013, 2016-2017, 2020 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""Convert terminal output with (some) ANSI escape sequences to HTML.
   Also escape any HTML special characters.

   At this time, most attribute ANSI escape sequences are converted, e.g.:

     ESC [ 0 m:         Reset attributes.
     ESC [ 41 m:        Background-color red.
     ESC [ 32 m:        Foreground-color red.
   """
import re

import html

HTML_STYLE = {
    1: "font-weight:bold",
    3: "font-style:italic",
    4: "text-decoration:underline",
    9: "text-decoration:line-through",
    22: "font-weight:normal",
    23: "font-style:normal",
    24: "text-decoration:none",
    29: "text-decoration:none",
    30: "color:black",
    31: "color:red",
    32: "color:green",
    33: "color:yellow",
    34: "color:blue",
    35: "color:magenta",
    36: "color:cyan",
    37: "color:white",
    40: "background-color:black",
    41: "background-color:red",
    42: "background-color:green",
    43: "background-color:yellow",
    44: "background-color:blue",
    45: "background-color:magenta",
    46: "background-color:cyan",
    47: "background-color:white",
}


def convert(msg):
    msg = html.escape(msg, quote=False)
    result = ""

    pattern = re.compile(r"\033\[(\d+)(;(\d+))*m")
    span_count = 0
    while True:
        m = pattern.search(msg)
        if not m:
            result += msg
            break
        result += msg[0 : m.start()]
        code = int(m.group(1))
        if code == 0:
            result += span_count * "</span>"
            span_count = 0
        elif code in HTML_STYLE:
            result += '<span style="%s">' % HTML_STYLE[code]
            span_count += 1
        msg = msg[m.end(0) :]
    return result + span_count * "</span>"


def selftest():
    test = "This is in \033[32mgreen\033[0m foreground."
    print("convert('%s') -> '%s'" % (test, convert(test)))
    test = (
        "This is in \033[41mred\033[0m background.  This \033[41mtoo\033[0m."
    )
    print("convert('%s') -> '%s'" % (test, convert(test)))
    test = (
        "\033[1mBold and \033[3mitalic and \033[4munderline\033[0m and none."
    )
    print("convert('%s') -> '%s'" % (test, convert(test)))


if __name__ == "__main__":
    selftest()



#
# Copyright (c) 2018-2021 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#  All rights reserved.
#
#  This code is the property of eGauge Systems LLC and may not be
#  copied, modified, or disclosed without any prior and written
#  permission from eGauge Systems LLC.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
import argparse
import sys

from intelhex import IntelHex

import egauge.ctid as CTid

CTid_table_addr = 0x3C0  # table goes in last 64 bytes

MAX_SERNO = 0xFFFFFF


def is_abbrev(abbrev, full):
    if not abbrev:
        return False
    if len(abbrev) > len(full):
        return False
    return abbrev == full[0 : len(abbrev)]


def auto_int(x):
    """Accept base-prefixed decimal."""
    return int(x, 0)


def edge_mask(x):
    try:
        return int(x)
    except ValueError:
        pass
    if is_abbrev(x, "rising"):  # rising
        return 0x1
    if is_abbrev(x, "falling"):
        return 0x2
    if is_abbrev(x, "both"):
        return 0x3
    raise ValueError(
        'Edge mask must be one of "rising", "falling", or "both".'
    )


def cal_table_entry(string):
    pair = string.split(":")
    if len(pair) != 2:
        raise argparse.ArgumentTypeError("missing `:'")
    idx = float(pair[0])
    if idx not in table.cal_table.keys():
        allowed = ", ".join(("%g" % k) for k in table.cal_table)
        raise argparse.ArgumentTypeError(
            "Invalid index %s (must be one " "of %s)." % (idx, allowed)
        )
    val = pair[1].split("/")
    if len(val) != 2:
        raise argparse.ArgumentTypeError("missing `/'")
    v_adj = float(val[0])
    p_adj = float(val[1])
    return (idx, v_adj, p_adj)


def hexdump(title, data):
    print(title + ":", end="")
    c = 0
    for b in data:
        if c % 8 == 0:
            print("\n 0x%08x:" % c, end="")
        print(" 0x%02x" % b, end="")
        c += 1
    print("")


def encode_model_str(model_str):
    val = model_str.encode("utf-8")
    if len(val) > 4:
        print(
            "%s: model name `%s' is %u bytes long in UTF-8 "
            "but only up to 4 bytes are allowed."
            % (parser.prog, model_str, len(val)),
            file=sys.stderr,
        )
        sys.exit(1)
    while len(val) < 4:
        val += b"\0"
    return val


def main():
    global parser, table
    table = CTid.Table()

    parser = argparse.ArgumentParser(
        description="Encode CTid parameters.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "program_template",
        help="Filename of the program template.",
        nargs=1,
        default=None,
    )
    parser.add_argument(
        "-d", "--debug", nargs=1, type=int, default=[0], help="Debug level"
    )
    parser.add_argument(
        "-o",
        "--output-file",
        nargs=1,
        type=argparse.FileType("w"),
        default=[sys.stdout],
    )
    #
    # Common information:
    #
    parser.add_argument(
        "-S",
        "--sensor-type",
        nargs=1,
        default=["AC"],
        choices=CTid.SENSOR_TYPE_NAME,
        help="Sensor-type.",
    )
    parser.add_argument(
        "-M",
        "--manufacturer",
        nargs=1,
        default=["0"],
        help="Manufacturer name or ID.",
    )
    parser.add_argument(
        "-m",
        "--model",
        nargs=1,
        default=[""],
        help="Model name (up to 4 characters).",
    )
    parser.add_argument(
        "-n",
        "--serial-number",
        nargs=1,
        type=auto_int,
        default=[0],
        help="Serial number.",
    )
    parser.add_argument(
        "-r",
        "--r-source",
        nargs=1,
        type=float,
        default=[0],
        help="Source resistance in Ohms.",
    )
    parser.add_argument(
        "-l",
        "--r-load",
        nargs=1,
        type=float,
        default=[0],
        help="Rated load resistance in Ohms.",
    )
    parser.add_argument(
        "-V",
        "--table-version",
        nargs=1,
        type=int,
        default=[CTid.CTID_VERSION],
        help="Version of the CTid table to generate.",
    )
    #
    # CT Parameters:
    #
    parser.add_argument(
        "-a",
        "--calibration-entry",
        nargs=1,
        type=cal_table_entry,
        action="append",
        help="Calibration-entry of the form X:VA/PA, where "
        "X is the percentage of the rated current to which the "
        "entry applies, VA is the voltage-adjustment in %% and "
        "PA is the phase-adjustment in \u00b0.",
    )
    parser.add_argument(
        "-c",
        "--rated-current",
        nargs=1,
        type=float,
        default=[100],
        help="Rated current of CT in Ampère.",
    )
    parser.add_argument(
        "-i",
        "--manufacturer-info",
        nargs=1,
        type=auto_int,
        default=[0],
        help="Manufacturer-specific information.",
    )
    parser.add_argument(
        "-p",
        "--phase",
        nargs=1,
        type=float,
        default=[0],
        help="Phase at rated current in \u00b0.",
    )
    parser.add_argument(
        "-s",
        "--size",
        nargs=1,
        type=float,
        default=[0],
        help="Size of the sensor in millimeter.",
    )
    parser.add_argument(
        "-t",
        "--voltage-temp-coeff",
        nargs=1,
        type=float,
        default=[0],
        help="Voltage temperature coefficient in ppm/\u00b0C.",
    )
    parser.add_argument(
        "-T",
        "--phase-temp-coeff",
        nargs=1,
        type=float,
        default=[0],
        help="Phase temperature coefficient in m\u00b0/\u00b0C.",
    )
    parser.add_argument(
        "-v",
        "--rated-voltage",
        nargs=1,
        type=float,
        default=[1 / 3.0],
        help="Rated voltage of CT in Volt.",
    )
    parser.add_argument(
        "-b",
        "--bias-voltage",
        type=float,
        default=0,
        help="Output voltage when no current " "is flowing through the CD.",
    )
    #
    # Linear Parameters:
    #
    parser.add_argument("--scale", type=float, default=None, help="Scale.")
    parser.add_argument("--offset", type=float, default=None, help="Offset.")
    parser.add_argument(
        "--delay", type=float, default=None, help="Signal delay in μs."
    )
    parser.add_argument(
        "--unit",
        type=int,
        default=0,
        help="Physical unit code (0..%d)." % (len(CTid.SENSOR_UNITS) - 1),
    )
    #
    # NTC Temperature Parameters:
    #
    parser.add_argument(
        "--ntc-a", type=float, default=None, help="NTC A coefficient."
    )
    parser.add_argument(
        "--ntc-b", type=float, default=None, help="NTC B coefficient."
    )
    parser.add_argument(
        "--ntc-c", type=float, default=None, help="NTC C coefficient."
    )
    parser.add_argument(
        "--ntc-m", type=float, default=None, help="NTC parameter M."
    )
    parser.add_argument(
        "--ntc-n", type=float, default=None, help="NTC parameter N."
    )
    parser.add_argument(
        "--ntc-k", type=float, default=None, help="NTC parameter K."
    )
    #
    # Pulse Temperature Parameters:
    #
    parser.add_argument(
        "--threshold",
        type=float,
        default=None,
        help="Pulse threshold in Volts.",
    )
    parser.add_argument(
        "--hysteresis",
        type=float,
        default=None,
        help="Hysteresis for threhold in Volts.",
    )
    parser.add_argument(
        "--debounce-time",
        type=float,
        default=None,
        help="Debounce time for pulse in milli-seconds.",
    )
    parser.add_argument(
        "--edge-mask",
        type=edge_mask,
        default=None,
        help="Edge-mask.  May be an integer or one of "
        '"rising", "falling", or "both".',
    )

    args = parser.parse_args()

    table_version = args.table_version[0]

    table.mfg_id = CTid.get_mfg_id_for_name(args.manufacturer[0])
    table.model = args.model[0]
    table.size = args.size[0]
    table.serial_number = args.serial_number[0]
    table.sensor_type = CTid.get_sensor_type_id(args.sensor_type[0])
    table.r_source = args.r_source[0]
    table.r_load = args.r_load[0]
    # CT parameters:
    table.rated_current = args.rated_current[0]
    table.voltage_at_rated_current = args.rated_voltage[0]
    table.bias_voltage = args.bias_voltage
    table.phase_at_rated_current = args.phase[0]
    table.voltage_temp_coeff = args.voltage_temp_coeff[0]
    table.phase_temp_coeff = args.phase_temp_coeff[0]
    if args.calibration_entry is not None:
        for cal in args.calibration_entry:
            c = cal[0]
            idx = c[0]
            table.cal_table[idx][0] = c[1]
            table.cal_table[idx][1] = c[2]
    # Voltage and temperature parameters:
    if args.scale is not None:
        table.scale = args.scale
    if args.offset is not None:
        table.offset = args.offset
    if args.delay is not None:
        table.delay = args.delay
    if args.unit is not None:
        table.sensor_unit = args.unit
    # NTC parameters:
    if args.ntc_a is not None:
        table.ntc_a = args.ntc_a
    if args.ntc_b is not None:
        table.ntc_b = args.ntc_b
    if args.ntc_c is not None:
        table.ntc_c = args.ntc_c
    if args.ntc_m is not None:
        table.ntc_m = args.ntc_m
    if args.ntc_n is not None:
        table.ntc_n = args.ntc_n
    if args.ntc_k is not None:
        table.ntc_k = args.ntc_k
    # Pulse parameters:
    if args.threshold is not None:
        table.threshold = args.threshold
    if args.hysteresis is not None:
        table.hysteresis = args.hysteresis
    if args.debounce_time is not None:
        table.debounce_time = args.debounce_time
    if args.edge_mask is not None:
        table.edge_mask = args.edge_mask

    table.mfg_info = int(args.manufacturer_info[0])

    if args.debug[0] > 0:
        print("CTid params:\n\t", table)

    try:
        table_data = table.encode(table_version)
    except CTid.Error:
        print("%s: %s" % (parser.prog, sys.exc_info()[1]), file=sys.stderr)
        sys.exit(1)

    if args.debug[0] > 0:
        hexdump("table", bytes((CTid.START_SYM,)) + table_data)

    bitstream = CTid.bitstuff(table_data)

    if args.debug[0] > 0:
        hexdump("bit stream", bitstream)

    ihex = IntelHex(args.program_template[0])

    max_addr = max(ihex.todict().keys())
    if max_addr > CTid_table_addr:
        print(
            "%s: max address in program file 0x%x is >= table address 0x%x"
            % (parser.prog, max_addr, CTid_table_addr)
        )
        sys.exit(1)

    addr = CTid_table_addr
    ihex[addr] = len(bitstream)
    addr += 1
    for i, byte in enumerate(bitstream):
        ihex[addr + i] = byte

    ihex.write_hex_file(args.output_file[0])


parser = None
table = None
if __name__ == "__main__":
    main()



#
# Copyright (c) 2020 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""This module provides access to the eGauge Serial Number web
   service.

"""
import logging

import urllib.parse

from .. import json_api

from ..error import Error

log = logging.getLogger(__name__)


class SerialNumberError(Error):
    """Raised for Serial Number API errors."""


class SerialNumber:
    def __init__(self, auth=None):
        """Return a serial number object.  AUTH should be an authentication
        object that provides the credentials to access the service.
        Typically, this should be a TokenAuth object.

        """
        self.api_uri = "https://api.egauge.net/v1/serial-numbers/"
        self.auth = auth

    def _get(self, resource, **kwargs):
        """Issue GET request for serial-number resource RESOURCE and return
        the parsed JSON data or None if the request failed or returned
        invalid JSON data.  Additional keyword arguments are passed on
        to requests.get().

        """
        reply = json_api.get(self.api_uri + resource, auth=self.auth, **kwargs)
        log.debug("get[%s] = %s", resource, reply)
        return reply

    def _post(self, resource, json_data, **kwargs):
        """Issue POST request to serial-number resource RESOURCE and return
        parsed JSON reply or None if the request failed or returned
        invalid JSON data.  Additional keyword arguments are passed on
        to requests.post().

        """
        reply = json_api.post(
            self.api_uri + resource, json_data, auth=self.auth, **kwargs
        )
        log.debug("post[%s] = %s", resource, reply)
        return reply

    def allocate(self, model_name, serial=None):
        """Return and allocate the next available serial number for model
        MODEL_NAME.  Typically, the MODEL_NAME should be prefixed by
        the manufacturer name to ensure uniqueness.  For example, for
        eGauge model ETN100, MODEL_NAME would be 'eGauge-ETN100'.
        Once allocated, a serial number cannot be freed again so care
        should be taken to use all allocated numbers.

        If SERIAL is specified, allocate that specific serial number,
        if it is avaiable, or fail otherwise.  Specific serial-number
        allocation may not be allowed for all MODEL_NAMEs.

        On error, exception SerialNumberError is raised.

        """
        data = {"name": model_name}
        if serial is not None:
            data["serial"] = serial
        reply = self._post("models/allocate/", json_data=data)
        if reply is None:
            raise SerialNumberError("POST to allocate SN failed.", model_name)
        if not "serial" in reply:
            log.error(
                "Failed to allocate SN: model=%s, reply=%s.", model_name, reply
            )
            if "errors" in reply:
                raise SerialNumberError(
                    "Error during SN allocation.", model_name, reply["errors"]
                )
            raise SerialNumberError("SN allocation failed.", model_name)
        return int(reply["serial"])

    def get_models(self):
        """Return a list of all model names registered in the database.  Each
        object in the list has members id (internal database id of the
        model), name (the model name), and max_sn (the maximum
        serial-number).

        On error, exception SerialNumberError is raised.
        """
        reply = self._get("models/")
        if reply is None:
            raise SerialNumberError("Failed to get SN models.")
        return reply

    def create_model(self, model_name, max_sn):
        """Create a new model with name MODEL_NAME and maximum serial number
        MAX_SN. Returns True on success, False on error.

        """
        data = {"name": model_name, "max_sn": max_sn}
        reply = self._post("models/", data)
        if reply is None or "name" not in reply:
            return False
        return reply["name"] == model_name

    def get_devices(self, model_name=None, dev_filter=None):
        """Return a list of devices.  If MODEL_NAME is not None, only devices
        with that model name are returned.  If DEV_FILTER is not None,
        only devices matching the filter are returned.

        """
        resource = "devices/"
        if model_name is not None:
            quoted_model = urllib.parse.quote(model_name, safe="")
            resource += quoted_model + "/"

        if dev_filter is not None:
            resource += "?" + dev_filter

        reply = self._get(resource)
        if reply is None:
            raise SerialNumberError(
                "Failed to get metadata.", model_name, dev_filter
            )
        return reply

    def get_metadata(self, model_name, sn):
        """Return the JSON-blob metadata for model MODEL_NAME and
        serial-number SN.

        On error, exception SerialNumberError is raised.
        """
        quoted_model = urllib.parse.quote(model_name, safe="")
        resource = "devices/%s/%s/" % (quoted_model, sn)
        reply = self._get(resource)
        if reply is None:
            raise SerialNumberError(
                "Failed to get serial number record.", model_name, sn
            )
        if "metadata" not in reply or reply["metadata"] is None:
            log.warning("no metadata exists for SN %s.", sn)
            return {}
        return reply["metadata"]

    def set_metadata(self, model_name, sn, meta):
        """Set the metadata for model MODEL_NAME and serial-number SN to META.
        META must be a JSON-serializable object.

        NOTE: Using get_metadata()/set_metadata() to update parts of
              the JSON metadata is NOT atomic.  We recommended using
              higher-level synchronization (such as a ResourceLock) to
              ensure there are no conflicting updates.

        On error, exception SerialNumberError is raised.

        """
        quoted_model = urllib.parse.quote(model_name, safe="")
        resource = "devices/%s/%s/" % (quoted_model, sn)
        reply = self._post(resource, json_data={"metadata": meta})
        if reply is None:
            raise SerialNumberError("Failed to set metadata.", model_name, sn)
        if "errors" in reply:
            raise SerialNumberError(
                "Failed to save metadata.", model_name, sn, reply["errors"]
            )



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2017, 2019-2021 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#  All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
import math
import struct

import crcmod

from .bit_stuffer import BitStuffer

CTID_VERSION = 4  # version of CTid Specification this code implements
START_SYM = 0xFF

SENSOR_TYPE_AC = 0x0  # AC-only sensor
SENSOR_TYPE_DC = 0x1  # DC-capable sensor
SENSOR_TYPE_RC = 0x2  # differential-output sensor ("Rogowski Coil"...)
SENSOR_TYPE_VOLTAGE = 0x3  # DEPRECATED -- use SENSOR_TYPE_LINEAR instead
SENSOR_TYPE_LINEAR = 0x3
SENSOR_TYPE_TEMP_LINEAR = 0x4
SENSOR_TYPE_TEMP_NTC = 0x5
SENSOR_TYPE_PULSE = 0x6

# List of registered manufacturer IDs:
_MFG_ID = {
    0x0000: ["eGauge Systems LLC", "eGauge"],
    0x0001: ["Magnelab, Inc.", "Magnelab"],
    0x0002: ["Continental Control Systems LLC", "CCS"],
    0x0003: ["J&D Electronics", "J&D"],
    0x0004: ["Accuenergy, Inc.", "Accuenergy"],
}

# For backwards-compatibility:
MFG_ID = {}
for key, value in _MFG_ID.items():
    MFG_ID[key] = value[0]

SENSOR_TYPE_NAME = ["AC", "DC", "RC", "linear", "temp", "NTC", "pulse"]

SENSOR_UNITS = (
    ("V", "voltage"),
    ("A", "current"),
    ("Ah", "charge"),
    ("W", "power"),
    ("var", "reactive power"),
    ("VA", "apparent power"),
    ("Hz", "frequency"),
    ("Ω", "resistance"),
    ("°", "angle"),
    ("°C", "temperature"),
    ("%RH", "humidity"),
    ("Pa", "pressure"),
    ("g", "mass"),
    ("m/s", "velocity"),
    ("g/s", "mass flow"),
    ("m^3/s", "volumetric flow"),
    ("%", "percentage"),
    ("ppm", "parts-per-million"),
    ("", "air quality"),
    ("", "number"),
)

_START_SYM_ZC_COUNT = 16  # number of zero-crossings in start-symbol
_CLK_FREQ = 480  # nominal clock frequency of CTid carrier
_CLK_TOL = 50  # clock tolerance in %

_MAX_CLK_FREQ = math.ceil(_CLK_FREQ * (1.0 + (_CLK_TOL / 100)))
_MIN_CLK_FREQ = math.floor(_CLK_FREQ * (1.0 - (_CLK_TOL / 100)))
_MIN_CLK_PERIOD = 1e6 / _MAX_CLK_FREQ  # min. period in microseconds
_MAX_CLK_PERIOD = 1e6 / _MIN_CLK_FREQ  # max. period in microseconds

_START_SYM_MAX_DURATION = _START_SYM_ZC_COUNT / 2 * _MAX_CLK_PERIOD

_DEBUG = 0

crc8_rohc = crcmod.predefined.mkCrcFun("crc-8-rohc")
crc16_modbus = crcmod.predefined.mkCrcFun("modbus")

# Max. mantissa value for float12 format:
F12_MAX_MANTISSA = 0x1FF


class CRCError(Exception):
    def __init__(self, expected, got):
        super().__init__()
        self.expected = expected
        self.got = got

    def __str__(self):
        return "Expected CRC 0x%x but got 0x%x" % (self.expected, self.got)


class Error(Exception):
    pass


def get_mfg_id_for_name(name):
    """Return the manufacturer ID for the given name.  Name may be a
    prefix of the full manufacturer's name, as long as the prefix is
    unique.

    """
    match = None
    for key, value in MFG_ID.items():
        if len(value) >= len(name):
            if value[0 : len(name)] == name:
                if match:
                    raise Error(
                        "Manufacturer name %s is a prefix of multiple "
                        "registered names.  Please specify more "
                        "letters." % name
                    )
                match = key
    if match is not None:
        return match
    # int() raises ValueError if not a number (use 0x prefix for hex, etc.)
    return int(name, 0)


def get_sensor_type_id(name_or_number):
    lc_name = name_or_number.lower()
    for i, name in enumerate(SENSOR_TYPE_NAME):
        if name.lower() == lc_name:
            return i
    # int() raises ValueError if not a number (use 0x prefix for hex, etc.)
    return int(name_or_number, 0)


def get_sensor_type_name(sensor_type):
    if sensor_type < 0 or sensor_type >= len(SENSOR_TYPE_NAME):
        return None
    return SENSOR_TYPE_NAME[sensor_type]


def get_sensor_unit(unit_code):
    if 0 <= unit_code < len(SENSOR_UNITS):
        return SENSOR_UNITS[unit_code][0]
    return "?"


def get_sensor_unit_desc(unit_code):
    if 0 <= unit_code < len(SENSOR_UNITS):
        return SENSOR_UNITS[unit_code][1]
    return "?"


def time_diff(l, r):
    return l - r


def is_zero_crossing(curr, prev, mean):
    return (prev < mean) != (curr < mean)


class Window:
    def __init__(self, sampling_freq):
        self.sample_period = 1e6 / sampling_freq  # period in microseconds
        self.avg_len = math.ceil(_START_SYM_MAX_DURATION / self.sample_period)
        window_len = 2 * self.avg_len

        self.tolerance = None
        self.count = 0  # number of valid samples
        self.avg_count = 0  # number of averaged samples
        self.sum = 0  # sum of the valid samples
        self.mean = 0  # mean value of valid samples
        self.wr = 0  # index of next sample to be written
        self.prev_polarity = 0  # previous polarity
        self.ts = [None] * window_len
        self.val = [None] * window_len

    def enter_sample(self, ts, val):
        if self.avg_count < self.avg_len:
            self.avg_count += 1
        else:
            idx = self.wr + self.avg_len
            if idx >= len(self.val):
                idx -= len(self.val)
            self.sum -= self.val[idx]
        if self.count < len(self.val):
            self.count += 1
        self.sum += val
        self.val[self.wr] = val
        self.ts[self.wr] = ts
        self.wr = (self.wr + 1) % len(self.val)
        self.mean = self.sum / self.avg_count

    def at_zero_crossing(self):
        curr_val = self.val[(self.wr + len(self.val) - 1) % len(self.val)]
        curr_pol = curr_val >= self.mean
        prev_pol = self.prev_polarity
        self.prev_polarity = curr_pol
        return curr_pol != prev_pol

    def at_start_symbol(self):
        if self.count < 1:
            return None

        curr = (self.wr + len(self.val) - 1) % len(self.val)
        prev_val = self.val[curr]
        self.prev_polarity = prev_val >= self.mean
        end_ts = self.ts[curr]
        start_ts = end_ts - _START_SYM_MAX_DURATION
        total = 0
        avg_count = 0

        # See if we can find 16 zero-crossings within _START_SYM_MAX_DURATION:
        zc_ts = []
        for i in range(1, self.count):
            curr = (curr + len(self.val) - 1) % len(self.val)
            curr_ts = self.ts[curr]
            curr_val = self.val[curr]

            if time_diff(start_ts, curr_ts) > _MAX_CLK_PERIOD:
                return None

            total += curr_val
            avg_count += 1

            if not is_zero_crossing(curr_val, prev_val, self.mean):
                continue  # keep looking for zero-crossing
            prev_val = curr_val
            zc_ts.append(curr_ts)
            if len(zc_ts) >= _START_SYM_ZC_COUNT:
                first_ts = curr_ts
                break

        if len(zc_ts) < _START_SYM_ZC_COUNT:
            return None

        mean = total / avg_count

        # Verify that there were no zero-crossings for _START_SYM_MAX_DURATION
        # before the first edge of the start-symbol
        for i in range(i + 1, self.count):
            curr = (curr + len(self.val) - 1) % len(self.val)
            curr_ts = self.ts[curr]
            curr_val = self.val[curr]

            if is_zero_crossing(curr_val, prev_val, mean):
                return None
            if time_diff(first_ts, curr_ts) >= _START_SYM_MAX_DURATION:
                break
            prev_val = curr_val

        # Calculate average period based on the 7 full bit times the 16
        # zero-crossings cover (clock periods may be asymmetric):
        period = time_diff(zc_ts[1], zc_ts[_START_SYM_ZC_COUNT - 1]) / (
            _START_SYM_ZC_COUNT / 2 - 1
        )
        if period < _MIN_CLK_PERIOD or period > _MAX_CLK_PERIOD:
            return None

        self.tolerance = period / 4
        sample_tolerance = 1.2 * self.sample_period
        if self.tolerance < sample_tolerance:
            self.tolerance = sample_tolerance

        # Verify that each bit has the expected period:
        curr_ts = end_ts
        for i in range(0, _START_SYM_ZC_COUNT, 2):
            dt = time_diff(curr_ts, zc_ts[i + 1])
            if abs(dt - period) > self.tolerance:
                return None
            curr_ts = zc_ts[i + 1]

        # Now that we confirmed a start-symbol, commit to its mean value:
        self.sum = total
        self.avg_count = avg_count
        self.mean = mean
        return period


class ByteDecoder:
    def __init__(self):
        self.num_bits = 0
        self.val = 0
        self.edge_count = 0
        self.run_length = 0
        self.start_ts = None
        self.decoded_byte = None

    def reset(self):
        self.__init__()

    def timed_out(self, period, now):
        if self.start_ts is None:
            return False
        return time_diff(now, self.start_ts) > 4 * period

    def update(self, period, tolerance, ts):
        if self.start_ts is None:
            self.start_ts = ts
            return False

        if time_diff(ts, self.start_ts + period - tolerance) < 0:
            # got what looks like a 1-edge
            self.edge_count += 1
            return False

        bit = self.edge_count & 1
        if _DEBUG >= 2:
            print(
                "bit %d at 0x%08x (edge_count %d, start_ts 0x%08x)"
                % (bit, ts, self.edge_count, self.start_ts)
            )

        self.edge_count = 0
        self.start_ts = ts
        if self.run_length >= 7:
            # drop stuffer bit...
            self.run_length = 0
            return False

        if bit:
            self.run_length += 1
        else:
            self.run_length = 0

        self.val = (self.val << 1) | bit
        self.num_bits += 1

        if self.num_bits < 8:
            return False
        self.decoded_byte = self.val
        self.val = 0
        self.num_bits = 0
        return True

    def get_byte(self):
        return self.decoded_byte


class Decoder:
    def __init__(self, sampling_freq):
        self.w = Window(sampling_freq)
        self.bd = ByteDecoder()
        self.period = None

    def get_byte(self):
        return self.bd.get_byte()

    def add_sample(self, timestamp, value):
        self.w.enter_sample(timestamp, value)

        if self.period is None:
            self.period = self.w.at_start_symbol()
            if self.period is None:
                return 0
            self.bd.reset()
            return -1  # got a new start-symbol

        if self.bd.timed_out(self.period, timestamp):
            # check if implied last edge finishes a byte, if so, return it:
            ret = self.bd.update(self.period, self.w.tolerance, timestamp)
            self.period = None
            return ret

        if not self.w.at_zero_crossing():
            return 0
        if _DEBUG >= 3:
            print(
                "zero-crossing at 0x%08x (mean %d)" % (timestamp, self.w.mean)
            )
        return self.bd.update(self.period, self.w.tolerance, timestamp)


def mfg_name(ident):
    """Return the manufacturer name for a given manufacturer IDENT."""
    if ident in _MFG_ID:
        return _MFG_ID[ident][0]
    return None


def mfg_short_name(ident):
    """Return the manufacturer name for a given manufacturer IDENT."""
    if ident in _MFG_ID:
        return _MFG_ID[ident][1]
    return None


def get_mfg_id(ident):
    """For backwards compatibility.  Please use mfg_name() instead."""
    return mfg_name(ident)


def fix(val, num_bits, is_signed, name, unit, scale):
    f = round(val / scale)
    limit = 1 << num_bits
    min_val = -(limit / 2) if is_signed else 0
    max_val = +(limit / 2 - 1) if is_signed else limit - 1
    if f < min_val or f > max_val:
        unit_str = " %s" % unit if unit is not None else ""
        prec = -int(math.log10(scale))
        val_str = ("%." + str(prec) + "f") % (val)
        min_str = ("%." + str(prec) + "f") % ((min_val * scale))
        max_str = ("%." + str(prec) + "f") % ((max_val * scale))
        raise Error(
            "%s %s%s outside of range from %s to %s%s."
            % (name, val_str, unit_str, min_str, max_str, unit_str)
        )
    return f


def unfix(val, scale):
    return val * scale


def float12(val, name, unit):
    """Encode a floating point value in the range from 0 to 1e7 to the
    12-bit float format.

    """
    if val < 0 or val > 1e7:
        raise Error(
            "%s %s%s outside of range from 0 to 10,000,000%s"
            % (name, val, unit, unit)
        )
    exp = math.log10(val) if val >= 1.0 else 0
    exp = int(math.ceil(exp))
    mantissa = int(round((val / math.pow(10, exp)) * F12_MAX_MANTISSA))
    return (mantissa << 3) | exp


def unfloat12(val):
    """Decode a 12-bit float value to a floating point value in the range
    from 0 to 1e7.

    """
    mantissa = (val >> 3) & F12_MAX_MANTISSA
    exp = val & 0x7
    return mantissa / float(F12_MAX_MANTISSA) * math.pow(10, exp)


def decimal16(val, name, unit):
    """Encode a decimal value with a 11-bit signed mantissa and a
    power-of-ten exponent from -16..15 as a 16-bit integer.
    This covers the range from -1024e15 to +1023e15.

    """
    mantissa = 0
    exp = 0

    if val != 0.0:
        # Determine exponent that gives us an mantissa in the range
        # from -1024..-103 or 103..1023:
        while exp > -16 and -102.4 <= val <= 102.3:
            val *= 10.0
            exp -= 1
        mantissa = round(val)
        while mantissa < -1024 or mantissa > 1023:
            mantissa /= 10.0
            exp += 1
        if exp > 15:
            raise Error(
                "%s %s%s outside of range from -1024e15..1023e15%s"
                % (name, val, unit, unit)
            )
        mantissa = int(mantissa)
    return ((mantissa & 0x7FF) << 5) | (exp & 0x1F)


def undecimal16(val):
    """Decode decimal 16-bit value to a floating number number."""
    mantissa = (val >> 5) & 0x7FF
    if mantissa >= 0x400:
        mantissa -= 0x800
    exp = val & 0x1F
    if exp >= 0x10:
        exp -= 0x20
    return mantissa * math.pow(10, exp)


def check_table_data(raw_data):
    version = struct.unpack_from(">B", raw_data)[0]
    if version <= 1:
        length = 30
        if len(raw_data) < length + 1:  # excluding start symbol
            raise Error("Insufficient table data")
        exp_crc = crc8_rohc(raw_data[:length])
        got_crc = struct.unpack_from(">B", raw_data[length:])[0]
        if exp_crc != got_crc:
            raise CRCError(exp_crc, got_crc)
    elif version <= 3:
        length = 33
        if len(raw_data) < length + 2:  # excluding start symbol
            raise Error("Insufficient table data")
        exp_crc = crc16_modbus(raw_data[:length])
        got_crc = struct.unpack_from(">H", raw_data[length:])[0]
        if exp_crc != got_crc:
            raise CRCError(exp_crc, got_crc)
    else:
        length = 43
        if len(raw_data) < length + 2:  # excluding start symbol
            raise Error("Insufficient table data")
        exp_crc = crc16_modbus(raw_data[:length])
        got_crc = struct.unpack_from(">H", raw_data[length:])[0]
        if exp_crc != got_crc:
            raise CRCError(exp_crc, got_crc)
    return raw_data[:length]


class Table:
    """Objects of this class can be used to hold the contents of a CTid
    table in a Python-native format.  Use the encode() method to
    convert the table to a sequence of bytes, encoded as per CTid
    spedification or decode a sequence of bytes using the decode()
    method.

    """

    def __init__(self, data=None):
        if data is not None:
            data = check_table_data(data)

        self.encoding = False
        self.version = CTID_VERSION
        self.mfg_id = 0
        self.model = ""
        self.serial_number = 0
        self.sensor_type = SENSOR_TYPE_AC
        self.r_source = 0
        self.r_load = 0  # undefined

        self.size = 0
        self.rated_current = 0
        self.voltage_at_rated_current = 0
        self.phase_at_rated_current = 0
        self.voltage_temp_coeff = 0
        self.phase_temp_coeff = 0
        self.cal_table = {1.5: [0, 0], 5: [0, 0], 15: [0, 0], 50: [0, 0]}
        self.bias_voltage = 0
        self.reserved = 0
        self.mfg_info = 0
        self.scale = 0
        self.offset = 0
        self.delay = 0
        self.sensor_unit = 0
        self.threshold = 0
        self.hysteresis = 0
        self.debounce_time = 0
        self.edge_mask = 0
        self.ntc_a = 0
        self.ntc_b = 0
        self.ntc_c = 0
        self.ntc_m = 0
        self.ntc_n = 0
        self.ntc_k = 0

        # internal data:
        self._raw_data = None
        self._raw_offset = None

        if data is not None:
            self.decode(data)

    def __str__(self):
        mfg_id_str = get_mfg_id(self.mfg_id)
        if mfg_id_str is not None:
            mfg_id_str = '"' + mfg_id_str + '"'
        else:
            mfg_id_str = "0x%04x" % self.mfg_id

        sensor_type_str = get_sensor_type_name(self.sensor_type)
        if sensor_type_str is not None:
            sensor_type_str = '"' + sensor_type_str + '"'
        else:
            sensor_type_str = "0x%1x" % self.sensor_type

        ret = (
            "CTid {version=%d, sensor_type=%s, mfg_id=%s, "
            'model="%s", '
            % (self.version, sensor_type_str, mfg_id_str, self.model)
        )
        if (
            self.sensor_type >= SENSOR_TYPE_AC
            and self.sensor_type <= SENSOR_TYPE_RC
        ):
            ret += (
                "size=%.1fmm, serial=%u, current=%.1fA, voltage=%.6fV, "
                "bias=%.3fmV, "
                "phase=%.2f\u00b0, voltage_temp_coeff=%.0fppm/\u00b0C, "
                "phase_temp_coeff=%.1fm\u00b0/\u00b0C, reserved=0x%02x, "
                "mfg_info=0x%02x, Rs=%g, Rl=%g, "
                "cal={"
                "1.5%%: %+.2f%%/%+.2f\u00b0, "
                "5%%: %+.2f%%/%+.2f\u00b0, "
                "15%%: %+.2f%%/%+.2f\u00b0, "
                "50%%: %+.2f%%/%+.2f\u00b0}"
                % (
                    self.size,
                    self.serial_number,
                    self.rated_current,
                    self.voltage_at_rated_current,
                    1e3 * self.bias_voltage,
                    self.phase_at_rated_current,
                    self.voltage_temp_coeff,
                    self.phase_temp_coeff,
                    self.reserved,
                    self.mfg_info,
                    self.r_source,
                    self.r_load,
                    self.cal_table[1.5][0],
                    self.cal_table[1.5][1],
                    self.cal_table[5][0],
                    self.cal_table[5][1],
                    self.cal_table[15][0],
                    self.cal_table[15][1],
                    self.cal_table[50][0],
                    self.cal_table[50][1],
                )
            )
        elif self.sensor_type == SENSOR_TYPE_LINEAR:
            unit = get_sensor_unit(self.sensor_unit)
            ret += "scale=%g%s/V offset=%g%s delay=%gμs" % (
                self.scale,
                unit,
                self.offset,
                unit,
                self.delay,
            )
        elif self.sensor_type == SENSOR_TYPE_TEMP_LINEAR:
            ret += "scale=%g°C/V offset=%g°C" % (self.scale, self.offset)
        elif self.sensor_type == SENSOR_TYPE_TEMP_NTC:
            ret += "A=%g B=%g C=%g M=%g N=%g K=%g" % (
                self.ntc_a,
                self.ntc_b,
                self.ntc_c,
                self.ntc_m,
                self.ntc_n,
                self.ntc_k,
            )
        elif self.sensor_type == SENSOR_TYPE_PULSE:
            ret += "threshold=%g±%gV debounce=%ums edge=0x%x" % (
                self.threshold,
                self.hysteresis,
                self.debounce_time,
                self.edge_mask,
            )
        else:
            pass
        ret += "}"
        return ret

    def m_u2(self, name, scale=1, unit=None):
        fmt = ">B"
        if self.encoding:
            val = fix(self.__dict__[name], 2, False, name, unit, scale)
            self._raw_data += struct.pack(fmt, val << 6)
        else:
            val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
            self._raw_offset += struct.calcsize(fmt)
            self.__dict__[name] = unfix(val >> 6, scale)

    def m_u8(self, name, scale=1, unit=None):
        fmt = ">B"
        if self.encoding:
            val = fix(self.__dict__[name], 8, False, name, unit, scale)
            self._raw_data += struct.pack(fmt, val)
        else:
            val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
            self._raw_offset += struct.calcsize(fmt)
            self.__dict__[name] = unfix(val, scale)

    def m_s8(self, name, scale=1, unit=None, idx1=None, idx2=None):
        fmt = ">b"
        if self.encoding:
            if idx1 is not None and idx2 is not None:
                val = fix(
                    self.__dict__[name][idx1][idx2], 8, True, name, unit, scale
                )
            else:
                val = fix(self.__dict__[name], 8, True, name, unit, scale)
            self._raw_data += struct.pack(fmt, val)
        else:
            val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
            self._raw_offset += struct.calcsize(fmt)
            if idx1 is not None and idx2 is not None:
                self.__dict__[name][idx1][idx2] = unfix(val, scale)
            else:
                self.__dict__[name] = unfix(val, scale)

    def m_u16(self, name, scale=1, unit=None):
        fmt = ">H"
        if self.encoding:
            val = fix(self.__dict__[name], 16, False, name, unit, scale)
            self._raw_data += struct.pack(fmt, val)
        else:
            val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
            self._raw_offset += struct.calcsize(fmt)
            self.__dict__[name] = unfix(val, scale)

    def m_s16(self, name, scale=1, unit=None):
        fmt = ">h"
        if self.encoding:
            val = fix(self.__dict__[name], 16, True, name, unit, scale)
            self._raw_data += struct.pack(fmt, val)
        else:
            val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
            self._raw_offset += struct.calcsize(fmt)
            self.__dict__[name] = unfix(val, scale)

    def m_d16(self, name, unit):
        fmt = ">H"
        if self.encoding:
            val = self.__dict__[name]
            self._raw_data += struct.pack(fmt, decimal16(val, name, unit))
        else:
            val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
            self._raw_offset += struct.calcsize(fmt)
            self.__dict__[name] = undecimal16(val)

    def m_s12(self, name, scale=1, unit=None):
        fmt = ">H"
        if self.encoding:
            s12 = fix(self.__dict__[name], 12, True, name, unit, scale)
            if s12 < 0:
                s12 &= 0xFFF
            self._raw_data += struct.pack(fmt, s12 << 4)
        else:
            u16 = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
            self._raw_offset += struct.calcsize(fmt)
            s12 = (u16 >> 4) & 0x0FFF
            if s12 >= 0x800:
                s12 -= 0x1000
            self.__dict__[name] = unfix(s12, scale)

    def m_u4_s12(self, name4, scale4, unit4, name12, scale12=1, unit12=None):
        fmt = ">H"
        if self.encoding:
            u4 = fix(self.__dict__[name4], 4, False, name4, unit4, scale4)
            s12 = fix(self.__dict__[name12], 12, True, name12, unit12, scale12)
            if s12 < 0:
                s12 &= 0xFFF
            self._raw_data += struct.pack(fmt, (u4 << 12) | s12)
        else:
            u16 = struct.unpack_from(fmt, self._raw_data, self._raw_offset)[0]
            self._raw_offset += struct.calcsize(fmt)
            self.__dict__[name4] = unfix((u16 >> 12) & 0xF, scale4)
            s12 = u16 & 0x0FFF
            if s12 >= 0x800:
                s12 -= 0x1000
            self.__dict__[name12] = unfix(s12, scale12)

    def m_f12_f12(self, name1, unit1, name2=1, unit2=None):
        fmt = ">BH"
        if self.encoding:
            f1 = float12(self.__dict__[name1], name1, unit1)
            f2 = float12(self.__dict__[name2], name2, unit2)
            u24 = (f1 << 12) | f2
            self._raw_data += struct.pack(fmt, u24 >> 16, u24 & 0xFFFF)
        else:
            val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)
            u24 = (val[0] << 16) | val[1]
            self._raw_offset += struct.calcsize(fmt)
            self.__dict__[name1] = round(unfloat12((u24 >> 12) & 0xFFF))
            self.__dict__[name2] = round(unfloat12((u24 >> 0) & 0xFFF))

    def m_u24(self, name, scale=1, unit=None):
        fmt = ">BH"
        if self.encoding:
            val = fix(self.__dict__[name], 24, False, name, unit, scale)
            self._raw_data += struct.pack(
                fmt, (val >> 16) & 0xFF, val & 0xFFFF
            )
        else:
            val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)
            val = (val[0] << 16) | val[1]
            self._raw_offset += struct.calcsize(fmt)
            self.__dict__[name] = unfix(val, scale)

    def m_f32(self, name):
        fmt = ">f"
        if self.encoding:
            self._raw_data += struct.pack(fmt, self.__dict__[name])
        else:
            val = struct.unpack_from(fmt, self._raw_data, self._raw_offset)
            self._raw_offset += struct.calcsize(fmt)
            self.__dict__[name] = val[0]

    def m_utf8_4(self, name):
        if self.encoding:
            val = self.__dict__[name].encode("utf-8")
            if len(val) > 4:
                raise Error(
                    "%s `%s' is %u bytes long in UTF-8 "
                    "but only up to 4 bytes are allowed."
                    % (name, self.__dict__[name], len(val))
                )
            while len(val) < 4:
                val += b"\0"
            self._raw_data += val
        else:
            raw = b""
            for i in range(self._raw_offset, self._raw_offset + 4):
                if self._raw_data[i] == 0:
                    break
                raw += bytearray((self._raw_data[i],))
            self._raw_offset += 4
            self.__dict__[name] = raw.decode()

    def m_utf8_8(self, name):
        if self.encoding:
            val = self.__dict__[name].encode("utf-8")
            if len(val) > 8:
                raise Error(
                    "%s `%s' is %u bytes long in UTF-8 "
                    "but only up to 8 bytes are allowed."
                    % (name, self.__dict__[name], len(val))
                )
            while len(val) < 8:
                val += b"\0"
            self._raw_data += val
        else:
            raw = b""
            for i in range(self._raw_offset, self._raw_offset + 8):
                if self._raw_data[i] == 0:
                    break
                raw += bytearray((self._raw_data[i],))
            self._raw_offset += 8
            self.__dict__[name] = raw.decode()

    def marshall_params_ct(self):
        """Current Transducer Parameters"""
        self.m_u16("rated_current", 0.1, "A")
        self.m_u16("voltage_at_rated_current", 10e-6, "V")
        self.m_u16("size", 0.1, "mm")
        self.m_s12("phase_at_rated_current", 0.01, "\u00b0")
        self.m_s8("voltage_temp_coeff", 5, "ppm/\u00b0C")
        self.m_s8("phase_temp_coeff", 0.5, "m\u00b0/\u00b0C")
        for k in sorted(self.cal_table.keys()):
            self.m_s8("cal_table", 0.02, "%", idx1=k, idx2=0)
            self.m_s8("cal_table", 0.02, "\u00b0", idx1=k, idx2=1)
        self.m_s16("bias_voltage", 1e-6, "V")

    def marshall_params_basic_linear(self):
        """Basic linear Parameters (for linear temp. and generic linear sensors)."""
        self.m_f32("scale")
        self.m_f32("offset")

    def marshall_params_linear(self):
        """Linear Parameters."""
        self.marshall_params_basic_linear()
        self.m_s16("delay", 0.01, "μs")
        self.m_u16("sensor_unit")

    def marshall_params_ntc(self):
        """NTC Parameters."""
        self.m_f32("ntc_a")
        self.m_f32("ntc_b")
        self.m_f32("ntc_c")
        self.m_f32("ntc_m")
        if self.version == 3:
            self.m_d16("ntc_r1", "Ω")
        else:
            self.m_f32("ntc_n")
            self.m_f32("ntc_k")

    def marshall_params_pulse(self):
        """Pulse Parameters."""
        self.m_u16("threshold", 10e-6, "V")
        self.m_u16("hysteresis", 10e-6, "V")
        self.m_u8("debounce_time")
        self.m_u2("edge_mask")

    def marshall_params(self):
        if (
            self.sensor_type >= SENSOR_TYPE_AC
            and self.sensor_type <= SENSOR_TYPE_RC
        ):
            self.marshall_params_ct()
        elif self.sensor_type == SENSOR_TYPE_LINEAR:
            self.marshall_params_linear()
        elif self.sensor_type == SENSOR_TYPE_TEMP_LINEAR:
            self.marshall_params_basic_linear()
        elif self.sensor_type == SENSOR_TYPE_TEMP_NTC:
            self.marshall_params_ntc()
        elif self.sensor_type == SENSOR_TYPE_PULSE:
            self.marshall_params_pulse()

        max_size = 43 if self.version >= 4 else 33

        if len(self._raw_data) < max_size:
            self._raw_data += (max_size - len(self._raw_data)) * b"\0"
        elif len(self._raw_data) > max_size:
            raise Error("CTid table too big", len(self._raw_data), max_size)

    def marshall(self):
        self.m_u8("version")
        self.m_u16("mfg_id")
        if self.version >= 4:
            # v4 or newer
            self.m_utf8_8("model")
            self.m_u8("reserved")
            self.m_u24("serial_number")
            self.m_u8("sensor_type")
            self.m_f12_f12("r_source", "Ω", "r_load", "Ω")
            self.marshall_params()
        else:
            self.m_utf8_4("model")
            if self.version == 3:
                # v3
                self.m_u24("serial_number")
                self.m_u8("sensor_type")
                self.m_f12_f12("r_source", "Ω", "r_load", "Ω")
                self.m_u8("reserved")
                self.marshall_params()
            else:
                # v1 or v2
                self.m_u16("size", 0.1, "mm")
                self.m_u24("serial_number")
                self.m_u16("rated_current", 0.1, "A")
                self.m_u16("voltage_at_rated_current", 10e-6, "V")
                self.m_u4_s12(
                    "sensor_type",
                    1,
                    None,
                    "phase_at_rated_current",
                    0.01,
                    "\u00b0",
                )
                self.m_s8("voltage_temp_coeff", 5, "ppm/\u00b0C")
                self.m_s8("phase_temp_coeff", 0.5, "m\u00b0/\u00b0C")
                for k in sorted(self.cal_table.keys()):
                    self.m_s8("cal_table", 0.02, "%", idx1=k, idx2=0)
                    self.m_s8("cal_table", 0.02, "\u00b0", idx1=k, idx2=1)
                self.m_u8("reserved")
                self.m_u8("mfg_info")
                if self.version > 1:
                    self.m_f12_f12("r_source", "Ω", "r_load", "Ω")

    def encode(self, version=CTID_VERSION):
        """Encode the table contents and store is as a sequence of bytes in
        property ``raw_data''.

        """
        self.encoding = True
        self._raw_data = b""
        self._raw_offset = 0
        self.version = version
        if version < 2 and (self.r_source != 0.0 or self.r_load != 0.0):
            raise Error(
                "Unable to encode r_source and r_load in "
                "CTid table version %u." % version
            )
        try:
            self.marshall()
        except KeyError as e:
            raise Error("Required parameter missing from CTid table.") from e
        except struct.error as e:
            raise Error("CTid table parameter has an invalid value.") from e
        if self.version == 1:
            # v1 used an 8-bit CRC
            self._raw_data += struct.pack(">B", crc8_rohc(self._raw_data))
            assert len(self._raw_data) == 31  # excluding start symbol
        else:
            self._raw_data += struct.pack(">H", crc16_modbus(self._raw_data))
            if self.version < 4:
                # v2-3
                assert len(self._raw_data) == 35  # excluding start symbol
            else:
                # v4 and on...
                assert len(self._raw_data) == 45  # excluding start symbol
        return self._raw_data

    def decode(self, raw_data):
        """Decode the sequence of bytes given by raw_data and update the Table
        data with the decoded contents.  Note: raw_data should not
        have the start-symbol nor the CRC at the end.  Also, the CRC
        should have been checked and confirmed to be good before
        calling this routine.

        """
        self.encoding = False
        self._raw_data = raw_data
        self._raw_offset = 0
        self.marshall()


def bitstuff(data):
    """Apply bit-stuffing to the data and return the result, prefixed with
    the start symbol.

    """
    bs = BitStuffer(bytes((START_SYM,)))
    for b in data:
        bs.append(b)
    return bs.get_output()


def unstuff(bitstream):
    """Remove bit-stuffing from the data with the start symbol removed."""
    if bitstream[0] != START_SYM:
        raise Error("Bitstream missing start symbol.", bitstream[0], START_SYM)

    data = bytearray()
    run_length = 0
    byte = 0x00
    num_bits = 0
    for b in bitstream[1:]:
        mask = 0x80
        while mask != 0:
            if run_length >= 7:
                # drop a stuffer bit
                run_length = 0
            else:
                if b & mask:
                    byte |= 1 << (7 - num_bits)
                    run_length += 1
                else:
                    run_length = 0
                num_bits += 1
                if num_bits >= 8:
                    data.append(byte)
                    byte = 0
                    num_bits = 0
            mask >>= 1
    return data



#!/usr/bin/env python3
#
#   Copyright (c) 2021-2022 eGauge Systems LLC
#     1644 Conestoga St, Suite 2
#     Boulder, CO 80301
#     voice: 720-545-9767
#     email: dave@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
# pylint: disable=no-name-in-module
#
"""This module provides a helper function to ask for eGauge cloud API
credentials (aka eGuard credentials) via a popup dialog.  To use it,
pass the ask method to the TokenAuth's ask parameter after partially
evaluating it to bind the parent Qt window to "parent":

from functools import partial

import egauge.webapi.cloud.credentials

    # this `self' is a Pyside2 Ui_MainWindow:
    auth = webapi.auth.TokenAuth(ask=partial(credentials.ask, self))
    self.sn_api = webapi.cloud.SerialNumber(auth=auth)

"""
# pylint: disable=too-few-public-methods

import getpass

try:
    from PySide2.QtWidgets import QDialog
    from .gui.credentials_dialog import Ui_Credentials_Dialog

    have_pyside2 = True
except ModuleNotFoundError:
    have_pyside2 = False

    class QDialog:
        """Dummy class for use by Credentials_Dialog."""

    class Ui_Credentials_Dialog:
        """Dummy class for use by Credentials_Dialog."""


class LoginCanceled(Exception):
    """Raised when the user cancels a login request."""


class CredentialsManager:
    def __init__(self, gui_parent=None):
        """Create a credentials manager.  The task of this manager is mainly
        to track if a previous login failed.  The user of this object
        should set the previous_login_failed member to False after the
        credentials have been used successfully.

        If GUI_PARENT is not None, it must be the QT5 (PySide2) parent
        window to use for the dialog.  If it is None, the credentials
        will be requested via standard I/O (getpass).

        """
        self.parent = gui_parent if have_pyside2 else None
        self.previous_login_failed = False

    def ask(self):
        """Ask for the username, password, and optional token for an eGauge
        cloud API account (eGuard account).  Returns a tuple
        containing a username and password or raises LoginCanceled if
        the user presses the "Cancel" button.

        """
        if self.parent:
            dialog = CredentialsDialog(self.parent, self.previous_login_failed)
            dialog.exec_()
            if not dialog.accepted:
                raise LoginCanceled()
            self.previous_login_failed = True
            return (dialog.username, dialog.password + dialog.token)

        fail_msg = ""
        if self.previous_login_failed:
            fail_msg = "Login failed.  "
        print(fail_msg + "Please enter eGuard credentials.")
        try:
            usr = input("Username: ")
            pwd = getpass.getpass(prompt="Password[+token]: ")
        except (KeyboardInterrupt, EOFError) as e:
            raise LoginCanceled from e
        self.previous_login_failed = True
        return [usr, pwd]


class CredentialsDialog(QDialog, Ui_Credentials_Dialog):
    def __init__(self, parent, failed):
        self.accepted = False
        self.username = None
        self.password = None
        self.token = ""
        super().__init__(parent)
        self.setupUi(self)
        if failed:
            prompt = "Login failed. " + self.prompt_label.text()
            self.prompt_label.setText(prompt)
        self.username_lineEdit.setFocus()

    def exec_(self):
        self.accepted = False
        super().exec_()

    def accept(self):
        super().accept()
        self.accepted = True
        self.username = self.username_lineEdit.text()
        self.password = self.password_lineEdit.text()
        self.token = self.token_lineEdit.text()


# Aliases for backwards-compatibility.  Please use CredentialsManager
# and CredentialsDialog in new code.
Credentials_Manager = CredentialsManager
Credentials_Dialog = CredentialsDialog



import os
import setuptools

with open(os.path.join(os.path.dirname(__file__), "README.rst")) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setuptools.setup(
    name="egauge-python",
    version="0.5.9",
    packages=setuptools.find_namespace_packages(include="egauge.*"),
    install_requires=[
        "crcmod",
        "intelhex",
        "wheel",
        "pexpect",
        "requests>=2.4.2",
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": ["ctid-encoder = egauge.ctid.encoder:main"]
    },
    license="MIT License",  # example license
    description=".",
    long_description=README,
    long_description_content_type="text/x-rst",
    url="https://bitbucket.org/egauge/python/",
    author="David Mosberger-Tang",
    author_email="davidm@egauge.net",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
)



#
# Copyright (c) 2020 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""This module provides a helper class for a row of time-stamped
register data."""
import copy
import json


class RegisterRow:
    """A row of register data contains a timestamp and the (signed 64-bit)
    values for that timestamp.  The timestamp is stored in member TS
    and the register values are stored in dictionary REGS, indexed by
    register name.

    """

    def __init__(self, ts, regs=None):
        self.ts = ts
        if regs is None:
            regs = {}
        self.regs = copy.copy(regs)

    def avg(self, regname):
        """Return the time-average of the register value."""
        return self.regs[regname] / self.ts

    def __sub__(self, subtrahend):
        """Subtract two register rows from each other and return the result.
        The SUBTRAHEND must have values for all of the registers in
        the minuend row.

        """
        ret = RegisterRow(self.ts, self.regs)
        ret.ts = float(ret.ts - subtrahend.ts)
        for name in self.regs:
            ret.regs[name] -= subtrahend.regs[name]
        return ret

    def __str__(self):
        return "%s: %s" % (self.ts, json.dumps(self.regs))



#
# Copyright (c) 2020 eGauge Systems LLC
#       1644 Conestoga St, Suite 2
#       Boulder, CO 80301
#       voice: 720-545-9767
#       email: davidm@egauge.net
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""Module to provide access to a device's JSON WebAPI."""

from .. import json_api

from ..error import Error


class DeviceError(Error):
    """Raised if for device related errors."""


class Device:
    """This class provides access to an eGauge device's JSON WebAPI.
    See "Web API Design" document for details."""

    def __init__(self, dev_uri, auth=None):
        """Return a device object that can be used to access the device a
        address DEV_URI.  An example DEV_URI would be
        "http://proto1.egaug.es".  AUTH should be an authentication object
        that provides the credentials to access the device.  Typically,
        this should be a JWTAuth object.

        """
        self.api_uri = dev_uri + "/api"
        self.auth = auth
        self._reg_idx = None
        self._reg_formula = None

    def get(self, resource, **kwargs):
        """Issue GET request for /api resource RESOURCE and return the parsed
        JSON data or None if the request failed or returned invalid
        JSON data.  Additional keyword arguments are passed on to
        requests.get().

        """
        return json_api.get(self.api_uri + resource, auth=self.auth, **kwargs)

    def put(self, resource, json_data, **kwargs):
        """Issue PUT request with JSON_DATA as body to /api resource RESOURCE
        and return parsed JSON reply or None if the request failed or
        returned invalid JSON data.  Additional keyword arguments are
        passed on to requests.put().

        """
        return json_api.put(
            self.api_uri + resource, json_data, auth=self.auth, **kwargs
        )

    def post(self, resource, json_data, **kwargs):
        """Issue POST request with JSON_DATA as body to /api resource RESOURCE
        and return parsed JSON reply or None if the request failed or
        returned invalid JSON data.  Additional keyword arguments are
        passed on to requests.post().

        """
        return json_api.post(
            self.api_uri + resource, json_data, auth=self.auth, **kwargs
        )

    def delete(self, resource, **kwargs):
        """Issue DELETE request for /api resource RESOURCE and return parsed
        JSON reply or None if the request failed or returned invalid
        JSON data.  Additional keyword arguments are passed on to
        requests.post().

        """
        return json_api.delete(
            self.api_uri + resource, auth=self.auth, **kwargs
        )

    def _fetch_reg_info(self):
        """Fetch register info, including type and virtual register
        formulas."""
        reply = self.get("/register", params={"virtual": "formula"})
        if reply is None or "registers" not in reply:
            raise DeviceError("Failed to fetch register info.", reply)
        self._reg_idx = {}
        self._reg_formula = {}
        for reg in reply["registers"]:
            self._reg_idx[reg["name"]] = reg["idx"]
            self._reg_formula[reg["name"]] = reg.get("formula")

    def reg_idx(self, regname):
        """Return the register index for the register with name REGNAME.  This
        information is cached in the device since it is relatively
        expensive to get (requires a separate call to /api/register).

        """
        if self._reg_idx is None:
            self._fetch_reg_info()
        return self._reg_idx[regname]

    def reg_formula(self, regname):
        """Return the register formula for the register with name REGNAME.
        This information is cached in the device since it is
        relatively expensive to get (requires a separate call to
        /api/register).

        """
        if self._reg_formula is None:
            self._fetch_reg_info()
        return self._reg_formula[regname]
